/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gnutls_plugin.h>

static int manage_error_mnode_gnutls(mnode_t *mnode, int s_ret, int err_code);

PLUGIN_DEPENDANCIES("tcp");

PLUGIN_GBL_VARS;

gnutls_certificate_credentials_t gnutls_cred;

gnutls_priority_t gnutls_cache;

gnutls_dh_params_t gnutls_dh;

cipher_t gnutls[] = {
	{
		-1,
		"ssl",
		setup_mnode_server_cipher_gnutls,
		setup_mnode_client_cipher_gnutls,
		deinit_mnode_cipher_gnutls,
		deinit_mnode_cipher_gnutls,
		recv_mnode_gnutls,
		send_mnode_gnutls
	},
	{-1, "", NULL, NULL, NULL, NULL, NULL, NULL }
};

static const command_t gnutls_commands[] = {
	{"cafile", CMD_NONE, cafile_cmd, ""},
	{"certnkey-files", CMD_NONE, certnkey_files_cmd, ""},
	{"crlfile", CMD_NONE, crlfile_cmd, ""},
	{"dhfile", CMD_NONE, dhfile_cmd, ""},
	{ "", CMD_NONE, NULL, "" }
};

plugin_type_t init(int plugin_id, in_plugin_handlers_t *iph)
{
	ptable_t *p_gcmd = NULL;

	PLUGIN_SET_VARS(plugin_id, iph);

	gnutls[0].plugin_id = plugin_id;

	p_gcmd = init_command_set(gnutls_commands);
	if (p_gcmd == NULL)
		return -1;

	if (gnutls_global_init() != GNUTLS_E_SUCCESS)
		return -1;

	if (gnutls_certificate_allocate_credentials(&gnutls_cred) != GNUTLS_E_SUCCESS)
		return -1;

	if (gnutls_priority_init (&gnutls_cache, "NORMAL", NULL) < 0)
		return -1;

	parse_configuration(mdevice_g->config_file, p_gcmd);

	pt_destroy(p_gcmd);

	log_msg(INFO_LVL, "Loaded GNUTLS plugin\n");

	return PLUGIN_CIPHER;
}

int unload(void)
{
	gnutls_global_deinit();

	log_msg(INFO_LVL, "GnuTLS deinitialized\n");

	return 0;
}

cipher_t *get_ciphers(void)
{
	return gnutls;
}

#if defined __GNUC__
#pragma GCC diagnostic ignored "-Wint-to-pointer-cast"
#endif
int setup_mnode_server_cipher_gnutls(mnode_t *mnode)
{
	int gnutls_ret;
	int retries = 0;
	gnutls_session_t session;

	gnutls_init(&session, GNUTLS_SERVER);

	gnutls_priority_set(session, gnutls_cache);

	gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, gnutls_cred);

	gnutls_certificate_server_set_request(session, GNUTLS_CERT_REQUEST);

	gnutls_transport_set_ptr(session, (gnutls_transport_ptr_t) mnode->sock);

	while ((gnutls_ret = gnutls_handshake(session)) != GNUTLS_E_SUCCESS)
	{
		DEBUG_MSG("GNUTLS server cipher setup error: %s\n", gnutls_strerror(gnutls_ret));

		if ((gnutls_error_is_fatal(gnutls_ret) != 0) || (retries > 4))
		{
			log_msg(ERROR_LVL, "%s\n", gnutls_strerror(gnutls_ret));
			gnutls_deinit(session);
			return -1;
		}

		retries++;
	}

	if (set_mnode_cipher_extent(mnode, session) < 0)
	{
		gnutls_deinit(session);
		return -1;
	}

	DEBUG_MSG("GNUTLS: mnode server session set up\n");

	return 0;
}
#if defined __GNUC__
#pragma GCC diagnostic error "-Wint-to-pointer-cast"
#endif

#if defined __GNUC__
#pragma GCC diagnostic ignored "-Wint-to-pointer-cast"
#endif
int setup_mnode_client_cipher_gnutls(mnode_t *mnode)
{
	int gnutls_ret;
	int retries = 0;
	gnutls_session_t session;

	gnutls_init(&session, GNUTLS_CLIENT);

	gnutls_priority_set(session, gnutls_cache);

	gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, gnutls_cred);

	gnutls_transport_set_ptr(session, (gnutls_transport_ptr_t) mnode->sock);

	while ((gnutls_ret = gnutls_handshake(session)) != GNUTLS_E_SUCCESS)
	{
		DEBUG_MSG("GNUTLS client cipher setup error: %s\n", gnutls_strerror(gnutls_ret));

		if ((gnutls_error_is_fatal(gnutls_ret) != 0) || (retries > 4))
		{
			log_msg(ERROR_LVL, "%s\n", gnutls_strerror(gnutls_ret));
			gnutls_deinit(session);
			close(mnode->sock);
			return -1;
		}

		retries++;
	}

	DEBUG_MSG("GNUTLS: mnode client session set up\n");

	if (set_mnode_cipher_extent(mnode, session) < 0)
	{
		gnutls_deinit(session);
		return -1;
	}

	return 0;
}
#if defined __GNUC__
#pragma GCC diagnostic error "-Wint-to-pointer-cast"
#endif

int deinit_mnode_cipher_gnutls(mnode_t *mnode)
{
	gnutls_session_t session = (gnutls_session_t) get_mnode_cipher_extent(mnode);

	if (session == NULL)
		return 0;

	gnutls_bye (session, GNUTLS_SHUT_WR);
	gnutls_deinit(session);

	set_mnode_cipher_extent(mnode, NULL);

	return 0;
}

void recv_mnode_gnutls(int sock, short event, void *arg)
{
	int in_bytes = 0;
	mnode_t *mnode = (mnode_t *)arg;
	int r_bytes = 0;

	DEBUG_MSG("GNUTLS Reading\n");

	/*Make compiler happy*/
	if ((sock != mnode->sock) ||  (event != EV_READ))
		return;
    /***/

	if (mnode->r_bytes == mdevice_g->buffer_size)
	{
		DEBUG_MSG("Gnutls - Input buffer full\n");
		return manage_packet(mnode);
	}

	r_bytes = 4096;
	if ((mdevice_g->buffer_size - mnode->r_bytes) < 4096)
		r_bytes = mdevice_g->buffer_size - mnode->r_bytes;

	in_bytes = gnutls_record_recv((gnutls_session_t) get_mnode_cipher_extent(mnode),
								  (void *) (mnode->r_buffer + mnode->r_bytes), r_bytes);

	if (in_bytes <= 0)
	{
		if (manage_error_mnode_gnutls(mnode, in_bytes, errno) < 0)
			remove_mnode(mdevice_g, mnode);
		return;
	}

	mnode->r_bytes += in_bytes;
	mnode->bytes_recv += in_bytes;

	manage_packet(mnode);
}

void send_mnode_gnutls(int sock, short event, void *arg)
{
	int out_bytes = 0;
	mnode_t *mnode = (mnode_t *)arg;
	int s_bytes = 0;

	/*Make compiler happy*/
	if ((sock != mnode->sock) || (event != EV_WRITE))
		return;
    /***/

	DEBUG_MSG("GNUTLS Sending\n");

	s_bytes = mnode->w_bytes;

	if (mnode->w_rests > 0)
		s_bytes = mnode->w_rests;

	out_bytes =  gnutls_record_send((gnutls_session_t) get_mnode_cipher_extent(mnode), 
									(const void *) mnode->w_buffer, s_bytes);

	if(out_bytes < 0)	
	{	
		mnode->w_rests = s_bytes;
		if (manage_error_mnode_gnutls(mnode, out_bytes, errno) < 0)
			remove_mnode(mdevice_g, mnode);
		return;
	}

	if (mnode->w_rests > 0)
		mnode->w_rests = 0;

	mnode->w_bytes -= out_bytes;
	mnode->bytes_send += out_bytes;

	if (mnode->w_bytes <= 0)
	{
		mnode->w_bytes = 0;
		event_del(&mnode->write_event);
	}
	else
		memmove(mnode->w_buffer, mnode->w_buffer + out_bytes, mnode->w_bytes);
}

static int manage_error_mnode_gnutls(mnode_t *mnode, int s_ret, int err_code)
{
	err_code = 0;

	DEBUG_MSG("GNUTLS Send/Recv Error Handling: %d - %s\n", s_ret, gnutls_strerror(s_ret));

	if (gnutls_error_is_fatal(s_ret))
	{
		DEBUG_MSG("GNUTLS Fatal error\n");
		return -1;
	}

	switch (s_ret)
	{
	case GNUTLS_E_REHANDSHAKE:
		if (gnutls_handshake((gnutls_session_t) get_mnode_cipher_extent(mnode)) < 0)
			return -1;
		break;
	case GNUTLS_E_INTERRUPTED:
	case GNUTLS_E_AGAIN:
		if (gnutls_record_get_direction((gnutls_session_t) get_mnode_cipher_extent(mnode)) == 1)
			event_add(&mnode->write_event, NULL);
		break;
	default:
		return -1;
	}

	return 0;
}
