/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <tcp.h>

static void mnode_server_cipherize(int fd, short event, void *arg);
static void mnode_server_authentify(int fd, short event, void *arg);
static void mnode_server_finalize(mnode_t *mnode);
static void finalize_connect_mnode(int fd, short event, void *arg);

PLUGIN_DEPENDANCIES(NULL);

PLUGIN_GBL_VARS;

connector_t tcp[]= {
	{
		-1,
		"tcp",
		connect_mnode_tcp,
		disconnect_mnode_tcp,
		setup_mserver_tcp,
		accept_mnode_tcp,
		close_mserver_tcp
	},
	{-1, "", NULL, NULL, NULL, NULL, NULL }
};

plugin_type_t init(int plugin_id, in_plugin_handlers_t *iph)
{
	PLUGIN_SET_VARS(plugin_id, iph);

	tcp[0].plugin_id = plugin_id;

	log_msg(INFO_LVL, "Loaded TCP plugin\n");

	return PLUGIN_CONNECTOR;
}

int unload(void)
{
	return 0;
}

connector_t *get_connectors(void)
{
	return tcp;
}

int setup_mserver_tcp(mdevice_t *mdevice, mserver_t *mserver)
{
	int flags;

	if (mserver == NULL)
		return 0;

	if ((mserver->sock = socket(mserver->address.sa_family, SOCK_STREAM, 0)) < 0)
	{
		ERR("setup_mserver_tcp()");
		return -1;
	}

	/*Reuse address*/
	flags = 1;
	if(setsockopt(mserver->sock, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(int)) < 0)
		ERR("setup_mserver_tcp()");

	/*Binding address and port*/
	((struct sockaddr_in *) &(mserver->address))->sin_port = htons(mserver->port);

	if (bind(mserver->sock, &(mserver->address), sizeof(struct sockaddr)) < 0) 
	{
		ERR("setup_mserver_tcp()");
		close(mserver->sock);
		return -1;
	}

	/*SET TCP OPTIMISATION*/
	flags = 1;
	if(setsockopt(mserver->sock, IPPROTO_TCP, TCP_NODELAY, &flags, sizeof(int)) < 0)
		ERR("setup_mserver_tcp()");

	/*flags = IPTOS_LOWDELAY;
	if(setsockopt(mserver->sock, IPPROTO_IP, IP_TOS, &flags, sizeof(int)) < 0)
		ERR("setup_mserver_tcp()");*/

	/*Listening*/
	if(listen(mserver->sock, SOMAXCONN) < 0)
	{
		ERR("setup_mserver_tcp()");
		close(mserver->sock);
		return -1;
	}

	DEBUG_MSG("TCP Server running on %s:%d\n", mserver->host, mserver->port);

	if (mdevice->mservers != NULL)
		mserver->next = mdevice->mservers;

	mdevice->mservers = mserver;

	/*Adding it in event base*/
	DEBUG_MSG("Adding TCP server event\n");
	event_set(&mserver->server_event, mserver->sock, EV_READ | EV_PERSIST, accept_mnode_tcp, mdevice);
	event_add(&mserver->server_event, NULL);

	return 0;
}

int close_mserver_tcp(mdevice_t *mdevice, mserver_t *mserver)
{
	if (mdevice == NULL || mserver == NULL)
		return -1;

	close(mserver->sock);

	return 0;
}

void accept_mnode_tcp(int sock, short event, void *arg)
{
	socklen_t addrlen;
	mnode_t *mnode;
	int flags;

	mdevice_t *mdevice = (mdevice_t *)arg;
	mserver_t *mserver = NULL;

	(void) event;

	DEBUG_MSG("A Node attemps a connection...\n");

	if((mnode = init_mnode()) == NULL)
	{
		log_msg(ERROR_LVL, "Can't accept new connections due to lack of memory!\n");
		return;
	}

	mnode->is_server = true;

	/*Retrieve mserver*/
	for (mserver = mdevice->mservers;
		 mserver != NULL && mserver->sock != sock;
		 mserver = mserver->next);

	addrlen = sizeof(struct sockaddr);
	if ((mnode->sock = accept(mserver->sock, &mnode->address, &addrlen)) < 0)
	{
		ERR("accept_mnode_tcp()");
		remove_mnode(mdevice_g, mnode);
		return;
	}

	if (mnode->address.sa_family == AF_INET)
	{
		mnode->port = htons(((struct sockaddr_in *) &mnode->address)->sin_port);
		inet_ntop(mnode->address.sa_family, &((struct sockaddr_in *) &mnode->address)->sin_addr, 
				  mnode->host, IP_HOSTNAME_MAX_LENGTH);
	}
	else
	{
		mnode->port = htons(((struct sockaddr_in6 *) &mnode->address)->sin6_port);
		inet_ntop(mnode->address.sa_family, &((struct sockaddr_in6 *) &mnode->address)->sin6_addr, 
				  mnode->host, IP_HOSTNAME_MAX_LENGTH);
	}

	/*SET TCP OPTIMISATION*/
	flags = 1;
	if(setsockopt(mnode->sock, IPPROTO_TCP, TCP_NODELAY, &flags, sizeof(int)) < 0)
		ERR("accept_mnode_tcp()");

	/*flags = IPTOS_LOWDELAY;
	if(setsockopt(mnode->sock, IPPROTO_IP, IP_TOS, &flags, sizeof(int)) < 0)
		ERR("accept_mnode_tcp()");*/

	mnode->connector_id = mserver->connector_id;
	mnode->cipher_id = mserver->cipher_id;
	mnode->authenticator_id = mserver->authenticator_id;

	DEBUG_MSG("Acceptation waiting to cipherize\n");
	event_set(&mnode->read_event, mnode->sock, EV_READ | EV_TIMEOUT, mnode_server_cipherize, mnode);
	event_add(&mnode->read_event, &mdevice_g->timeout);
}

int connect_mnode_tcp(mnode_t *mnode)
{
	int flags;
	int ret_code;

	if ((mnode->sock = socket(mnode->address.sa_family, SOCK_STREAM, 0)) < 0)
	{
		ERR("connect_mnode_tcp()");
		return -1;
	}

	/*Setting non blocking mode*/
	if ((flags = fcntl(mnode->sock, F_GETFL, 0)) < 0)
        flags = 0;

    if (fcntl(mnode->sock, F_SETFL, flags | O_NONBLOCK) < 0)
	{
		ERR("connect_mnode_tcp()");
		close(mnode->sock);
		return -1;
	}

	flags = 1;
	if(setsockopt(mnode->sock, IPPROTO_TCP, TCP_NODELAY, &flags, sizeof(int)) < 0)
		ERR("connect_mnode_tcp()");

	if (mnode->address.sa_family == AF_INET)
		((struct sockaddr_in *) &mnode->address)->sin_port = htons(mnode->port);
	else
		((struct sockaddr_in6 *) &mnode->address)->sin6_port = htons(mnode->port);

	ret_code = connect(mnode->sock, &(mnode->address), sizeof(struct sockaddr));
	if (ret_code < 0)
	{
		if (errno != EINPROGRESS)
		{
			ERR("connect_mnode_tcp()");
			close(mnode->sock);
			return -1;
		}

		DEBUG_MSG("Connection waiting to finalize\n");
		event_set(&mnode->write_event, mnode->sock, EV_WRITE | EV_TIMEOUT, finalize_connect_mnode, mnode);
		event_add(&mnode->write_event, &mdevice_g->timeout);
		return 0;
	}

	DEBUG_MSG("Connection worked immediately!\n");
	finalize_connect_mnode(mnode->sock, EV_WRITE, mnode);

	return 0;
}

void disconnect_mnode_tcp(mnode_t *mnode)
{
	close(mnode->sock);
	DEBUG_MSG("Node %s disconnected\n", mnode->host);
}

static void mnode_server_cipherize(int fd, short event, void *arg)
{
	int ret_code;
	mnode_t *mnode = (mnode_t *) arg;

	(void) fd;

	if (event & EV_TIMEOUT)
	{
		remove_mnode(mdevice_g, mnode);
		return;
	}

	setup_mnode_server_cipher(ret_code, mnode);
	if (ret_code < 0)
	{
		remove_mnode(mdevice_g, mnode);
		return;
	}

	DEBUG_MSG("Cipher set!\n");

	if (mnode->authenticator_id >= 0)
	{
		DEBUG_MSG("Connection waiting to authentify\n");
		event_set(&mnode->read_event, mnode->sock, EV_READ | EV_TIMEOUT, mnode_server_authentify, mnode);
		event_add(&mnode->read_event, &mdevice_g->timeout);
		return;
	}

	mnode_server_finalize(mnode);
}

static void mnode_server_authentify(int fd, short event, void *arg)
{
	int ret_code;
	mnode_t *mnode = (mnode_t *) arg;
	
	(void) fd;

	if (event & EV_TIMEOUT)
	{
		remove_mnode(mdevice_g, mnode);
		return;
	}

	setup_mnode_server_authentification(ret_code, mnode);
	if (ret_code < 0)
	{
		remove_mnode(mdevice_g, mnode);
		return;
	}

	DEBUG_MSG("Authenticator set!\n");

	mnode_server_finalize(mnode);
}

static void mnode_server_finalize(mnode_t *mnode)
{
	int flags;

	/*Setting non blocking mode*/
	if ((flags = fcntl(mnode->sock, F_GETFL, 0)) < 0)
        flags = 0;

    if (fcntl(mnode->sock, F_SETFL, flags | O_NONBLOCK) < 0)
	{
		ERR("mnode_server_finalize()");
		remove_mnode(mdevice_g, mnode);
		return;
	}

	log_msg(INFO_LVL, "A node has connected %s:%d\n", mnode->host, mnode->port);

	/*Set events and add read one*/
	event_set(&mnode->read_event, mnode->sock, EV_READ | EV_PERSIST, recv_mnode_h(mnode), mnode);
	event_set(&mnode->write_event, mnode->sock, EV_WRITE | EV_PERSIST, send_mnode_h(mnode), mnode);
	event_add(&mnode->read_event, NULL);

	post_accept(ph->callbacks[CB_POST_ACCEPT], mdevice_g, mnode);
}

static void finalize_connect_mnode(int fd, short event, void *arg)
{
	int flags;
	int ret_code;
	mnode_t *mnode = (mnode_t *) arg;

	(void) fd;

	if (event & EV_TIMEOUT)
	{
		remove_mnode(mdevice_g, mnode);
		return;
	}
	
	if ((flags = fcntl(mnode->sock, F_GETFL, 0)) < 0)
        flags = 0;
	
	if (flags & O_NONBLOCK)
		flags ^= O_NONBLOCK;

	if (fcntl(mnode->sock, F_SETFL, flags) < 0)
	{
		ERR("finalize_connect_mnode()");
		remove_mnode(mdevice_g, mnode);
		return;
	}

	setup_mnode_client_cipher(ret_code, mnode);
	if (ret_code < 0)
	{
		DEBUG_MSG("setup_mnode_client_cipher failed\n");
		remove_mnode(mdevice_g, mnode);
		return;
	}

	mnode->authenticator_id = mdevice_g->authenticator_id;

	setup_mnode_client_authentification(ret_code, mnode);
	if (ret_code < 0)
	{
		DEBUG_MSG("setup_mnode_client_authentification failed\n");
		remove_mnode(mdevice_g, mnode);
		return;
	}

	log_msg(INFO_LVL, "Connected to mnode %s:%d\n", mnode->host, mnode->port);

	/*Setting non blocking mode*/
	if ((flags = fcntl(mnode->sock, F_GETFL, 0)) < 0)
        flags = 0;

    if (fcntl(mnode->sock, F_SETFL, flags | O_NONBLOCK) < 0)
	{
		ERR("finalize_connect_mnode()");
		remove_mnode(mdevice_g, mnode);
		return;
	}

	/*Set events and add read one*/
	event_set(&mnode->read_event, mnode->sock, EV_READ | EV_PERSIST, recv_mnode_h(mnode), mnode);
	event_set(&mnode->write_event, mnode->sock, EV_WRITE | EV_PERSIST, send_mnode_h(mnode), mnode);
	event_add(&mnode->read_event, NULL);
	
	post_connect(ph->callbacks[CB_POST_CONNECT], mdevice_g, mnode);
}
