/*    Copyright (C) 2010  Tomasz Bursztyka
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <manage.h>

static void accept_mng_socket(int sock, short event, void *null);
static void authenticate_client(int sock, short event, void *null);
static void manage(int sock, short event, void *null);
static int error_client(short event, int s_ret, int err_code);
static void client_msg(log_level_t lvl, const char *format, ...);
static void send_client_msg(char *line);
static void recv_client_msg(int sock, char *buffer, int buff_len);

PLUGIN_DEPENDANCIES(NULL);

PLUGIN_GBL_VARS;

char mng_password[MNG_MAX_PASSWORD];
static int mng_server_socket = -1;
static int mng_client_socket = -1;
static struct event manage_server_event;
static struct event manage_client_event;

static const command_t mng_cmd[] = {
	{ "manage", CMD_INTERNAL, manage_cmd, "manage <password> <port> - run a simple TCP based interface \
on localhost to manage the whole system. Password is mandatory, port is optional (default is MNG_PORT_DEFAULT)" },
	{ "quit", CMD_EXTERNAL, quit_cmd, "quit - quit the management interface (client side)" },
	{ "disconnect", CMD_EXTERNAL, disconnect_cmd, "disconnect <mnode ip/uip> - disconnect selected mnode (targeted by its vpn ip/uip" },
	{ "list", CMD_EXTERNAL, list_cmd, "list <mnodes|mservers|connectors|cyphers|authenticators> - list one of this possible elements." }
};

plugin_type_t init(int plugin_id, in_plugin_handlers_t *iph)
{
	PLUGIN_SET_VARS(plugin_id, iph);
	unsigned short idx;

	for (idx = 0; idx < 4; idx++)
		if (add_command((command_t *) &mng_cmd[idx]) < 0)
			return -1;

	log_msg(INFO_LVL, "Loaded MANAGE plugin\n");

	return PLUGIN_DEFAULT;
}

int unload(void)
{
	unsigned short idx;

	client_disconnect();

	if (event_initialized(&manage_server_event))
		event_del(&manage_server_event);

	close(mng_server_socket);

	for (idx = 0; idx < 4; idx++)
		del_command((command_t *) &mng_cmd[idx]);

	return 0;
}

int setup_mng_socket(uint16_t port)
{
	struct sockaddr_in addr;
	int flags;

	if ((mng_server_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		ERR("setup_mng_socket()");
		return -1;
	}

	flags = 1;
	if(setsockopt(mng_server_socket, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(int)) < 0)
		ERR("setup_mng_socket()");

	flags = 1;
	if(setsockopt(mng_server_socket, IPPROTO_TCP, TCP_NODELAY, &flags, sizeof(int)) < 0)
		ERR("setup_mng_socket()");

	if ((flags = fcntl(mng_server_socket, F_GETFL, 0)) < 0)
        flags = 0;

    if (fcntl(mng_server_socket, F_SETFL, flags | O_NONBLOCK) < 0)
	{
		ERR("setup_mng_socket()");
		close(mng_server_socket);
		return -1;
	}

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	addr.sin_port = htons(port);

	if (bind(mng_server_socket, (struct sockaddr *) &addr, sizeof(struct sockaddr)) < 0)
	{
		ERR("setup_mng_socket()");
		close(mng_server_socket);
		return -1;
	}

	if(listen(mng_server_socket, 1) < 0)
	{
		ERR("setup_mng_socket()");
		close(mng_server_socket);
		return -1;
	}

	event_set(&manage_server_event, mng_server_socket, EV_READ|EV_PERSIST, accept_mng_socket, NULL);
	event_add(&manage_server_event, NULL);

	return 0;
}

void client_disconnect(void)
{
	if (event_initialized(&manage_client_event))
		event_del(&manage_client_event);

	close(mng_client_socket);
	mng_client_socket = -1;

	if (mng_server_socket > -1)
		event_add(&manage_server_event, NULL);
}

static void accept_mng_socket(int sock, short event, void *null)
{
	struct sockaddr mng_client_addr;
	socklen_t addrlen;
	int flags;
	struct timeval timeout;

	/*Make compiler happy*/
	null = NULL;
	if (event != EV_READ)
		log_msg(WARNING_LVL, "accept_mng_socket() called from wrong event");
	/**/

	addrlen = sizeof(struct sockaddr);

	if ((mng_client_socket = accept(sock, (struct sockaddr *) &mng_client_addr, &addrlen)) < 0)
	{
		ERR("accept_mng_socket()");
		return;
	}

	flags = 1;
	if(setsockopt(mng_client_socket, IPPROTO_TCP, TCP_NODELAY, &flags, sizeof(int)) < 0)
		ERR("accept_mng_socket()");

	if ((flags = fcntl(mng_client_socket, F_GETFL, 0)) < 0)
        flags = 0;

    if (fcntl(mng_client_socket, F_SETFL, flags | O_NONBLOCK) < 0)
	{
		ERR("accept_mng_socket()");
		close(mng_client_socket);
		return;
	}

	memset(&timeout, 0, sizeof(struct timeval));
	timeout.tv_sec = 10;

	event_once(mng_client_socket, EV_READ|EV_TIMEOUT, authenticate_client, NULL, &timeout);

	/*We don't want to get new connections until this one ends.*/
	event_del(&manage_server_event);

	send_client_msg("Password: ");
}

static void authenticate_client(int sock, short event, void *null)
{
	char password[MNG_MAX_PASSWORD];

	null = NULL;

	if (event == EV_TIMEOUT)
	{
		send_client_msg("*** TIMEOUT ***\n");
		client_disconnect();
		return;
	}

	recv_client_msg(sock, password, MNG_MAX_PASSWORD);

	if (mng_client_socket < 0)
		return;

	if (strlen(password) < 4)
	{
		client_disconnect();
		return;
	}

	if (password[strlen(password) - 2] == '\r')
		password[strlen(password) - 2] = '\0';

	if (strncmp(password, mng_password, MNG_MAX_PASSWORD) != 0)
	{
		DEBUG_MSG("Manage - Got wrong password %s\n", password);
		client_disconnect();
		return;
	}

	send_client_msg(GPL_TERMINAL_MESSAGE);
	send_client_msg("> ");

	event_set(&manage_client_event, sock, EV_READ|EV_PERSIST, manage, NULL);
	event_add(&manage_client_event, NULL);
}

static void manage(int sock, short event, void *null)
{
	char cmd_line[CMD_LINE_MAX_LENGTH];
	char **cmd;

	/*Make compiler happy*/
	null = NULL;
	if (event != EV_READ)
		log_msg(WARNING_LVL, "manage() called from wrong event");
	/**/

	recv_client_msg(sock, cmd_line, CMD_LINE_MAX_LENGTH);

	if (mng_client_socket < 0)
		return;

	if ((cmd = parse_command(cmd_line)) == NULL)
	{
		send_client_msg("> ");
		return;
	}

	log_set_redirection(client_msg);
	switch (interpret_command(cmd, CMD_EXTERNAL, NULL))
	{
	case CMD_UNKNOWN:
		send_client_msg("\nCommand unknown\n");
		break;
	case CMD_ERROR:
		send_client_msg("\nCommand error\n");
		break;
	case CMD_SUCCESS:
		send_client_msg("\nCommand success\n");
	default:
		break;
	}
	log_unset_redirection();

	free_cmd_line(cmd);

	send_client_msg("> ");
}

static int error_client(short event, int s_ret, int err_code)
{
	if (s_ret == 0)
		return -1;

	switch(err_code)
	{
	case EINTR:
	case EAGAIN:
		if (event == EV_READ)
			DEBUG_MSG("Manage: client read interrupted\n");
		if (event == EV_WRITE)
			DEBUG_MSG("Manage: client write interrupted\n");
		return 0;
	default:
		return -1;
	}

	return 0;
}

static void client_msg(log_level_t lvl, const char *format, ...)
{
	char line[CMD_LINE_MAX_LENGTH*2];
	va_list args;

	lvl = 0;

	memset(line, 0, CMD_LINE_MAX_LENGTH*2);

	va_start(args, format);
	vsnprintf(line, (CMD_LINE_MAX_LENGTH*2)-1, format, args);

	send_client_msg(line);
}

static void send_client_msg(char *line)
{
	int out_bytes = 0;

	out_bytes = send(mng_client_socket, line, strlen(line), 0);
	if (out_bytes <= 0)
		if (error_client(EV_WRITE, out_bytes, errno) < 0)
			client_disconnect();
}

static void recv_client_msg(int sock, char *buffer, int buff_len)
{
	int in_bytes;

	memset(buffer, 0, buff_len);

	in_bytes =  recv(sock, buffer, buff_len, 0);
	if (in_bytes <= 0)
		if (error_client(EV_READ, in_bytes, errno) < 0)
			client_disconnect();
}
